/*
 *	linux/fs/rbf/bitmap.c
 *	RBF File System - bitmap operations.
 *	
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"
#include "rbf_bitops.h"

#undef DEBUG
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

/*
 *	Return number of BLOCK containing bitmap entry for specified sector
 *	Also return total number of bits (clusters) in block at *lengthp
 *	& bit offset into block at *offsetp.
 */
static long rbf_bitmap_block(struct super_block *s, long sector, int *lengthp, int *offsetp)
{
	int cluster;
	
	if (sector >= s_su_nsect)
		return -1;
	
	cluster = s_su_mapoffs + (sector >> s_su_clus_bits);
	*lengthp = min(s_su_mapoffs + s_su_nclus - (cluster & -(s->s_blocksize * 8)), s->s_blocksize * 8);
	*offsetp = cluster & (s->s_blocksize * 8 - 1);
	
	DPRINTK("sector %08lx block %08lx (length %04x) offset %04x\n",
			sector, (cluster >> (s->s_blocksize_bits + 3)) + s_su_mapblk, *lengthp, *offsetp);
	
	return (cluster >> (s->s_blocksize_bits + 3)) + s_su_mapblk;
}



static inline void rbf_update_bitmap(struct super_block *s, struct buffer_head *bh)
{
	rbf_update_buffer(s->s_flags & MS_SYNCHRONOUS, bh);
}


int rbf_extend_allocation(struct super_block *s, sector_t sector, int nsect)
{
	int count = (nsect + s_su_clus - 1) >> s_su_clus_bits;	/* clusters */
	long mapblk;
	int length, offset, nextused, nclus;
	struct buffer_head *bh;
	
	DPRINTK("sector %08lx nsect %04x\n", (long unsigned int) sector, nsect);
	
	mapblk = rbf_bitmap_block(s, sector, &length, &offset);
	if (mapblk == -1)
		return 0;			/* end of media */
	
	bh = sb_bread(s, mapblk);
	if (bh == 0) {
		printk(KERN_ERR "RBF: error reading bitmap block %ld\n", mapblk);
		return -EIO;
	}
	lock_super(s);
	nextused = rbf_find_next_one_bit(bh->b_data, length, offset);
	
	DPRINTK("offset %d nextused %d\n", offset, nextused);
	
	nclus = min(nextused - offset, count);
	if (nclus <= 0) {			/* cannot extend - following cluster in use */
		unlock_super(s);
		brelse(bh);
		DPRINTK("returns 0\n");
		return 0;
	}
	nsect = nclus << s_su_clus_bits;
	rbf_set_bits(bh->b_data, offset, nclus);
	s_su_freesect -= nsect;
	rbf_update_bitmap(s, bh);
	unlock_super(s);
	brelse(bh);

	DPRINTK("returns %04x\n", nsect);
	return nsect;
}


int rbf_allocate(struct super_block *s, sector_t *sectorp, int nsect)
{
	int count = (nsect + s_su_clus - 1) >> s_su_clus_bits;	/* clusters */
	int length, offset, end;
	long mapblk, sector = 0, largest_sector = 0;
	int start, largest_start = 0;
	int size, largest_size = 0;
	struct buffer_head *bh;
	
	DPRINTK("nsect %04x (nclus=%d)\n", nsect, count);
	
	/*
	 *	search bitmap
	 */
	lock_super(s);
	while ((mapblk = rbf_bitmap_block(s, sector, &length, &offset)) != -1) {
		bh = sb_bread(s, mapblk);
		if (bh == 0) {
			printk(KERN_ERR "RBF: error reading bitmap block %ld\n", mapblk);
			goto next;
		}
		start = rbf_find_next_zero_bit(bh->b_data, length, offset);
		while (start < length) {
			end = rbf_find_next_one_bit(bh->b_data, length, start);
			DPRINTK("free area %d-%d\n", start, end);
			size = end - start;
			/*
			 *	Allocation strategy: use largest block in first
			 *	bitmap block containing a large enough block.
			 */
			if (size > largest_size) {
				largest_size = size;
				largest_start = start;
				largest_sector = sector;
			}
			start = rbf_find_next_zero_bit(bh->b_data, length, end);
		}
		if (largest_size > count) {
			start = largest_start;
			goto found;
		}
		brelse(bh);
	next:
		sector += (length - offset) << s_su_clus_bits;
	}
	if (largest_size) {
		sector = largest_sector;
		start = largest_start;
		count = largest_size;
	}
	else {
		unlock_super(s);
		return 0;			/* nothing free! */
	}
	
	mapblk = rbf_bitmap_block(s, sector, &length, &offset);
	bh = sb_bread(s, mapblk);
	if (bh == 0) {
		unlock_super(s);
		printk(KERN_ERR "RBF: error reading bitmap block %ld\n", mapblk);
		return -EIO;
	}
	
found:	
	rbf_set_bits(bh->b_data, start, count);
	nsect = count << s_su_clus_bits;
	s_su_freesect -= nsect;
	rbf_update_bitmap(s, bh);
	unlock_super(s);
	brelse(bh);

	*sectorp = sector + ((start - offset) << s_su_clus_bits);
	DPRINTK("returns %04x at %08lx\n", nsect, (long unsigned int) *sectorp);
	return nsect;
}


/*
 *	Deallocate from bitmap <nsect> sectors starting at <sector>
 */
void rbf_deallocate(struct super_block *s, sector_t sector, int nsect)
{
	int count = (nsect + s_su_clus - 1) >> s_su_clus_bits;	/* clusters */
	int mapblk, length, offset, nclus;
	struct buffer_head *bh;
	
	DPRINTK("%04x at %08lx\n", nsect, (long unsigned int) sector);
	
	if (sector & (s_su_clus - 1)) {
		printk(KERN_ERR "RBF: deallocate invalid cluster address %08lx\n", (long) sector);
		return;
	}
		
	while (count) {
		mapblk = rbf_bitmap_block(s, sector, &length, &offset);
		if (mapblk == -1) {
			printk(KERN_ERR "RBF: invalid sector # %ld\n", (long) sector);
			return;
		}
		nclus = min(count, length - offset);
		nsect = nclus << s_su_clus_bits;
		
		bh = sb_bread(s, mapblk);
		if (bh == 0) {
			printk(KERN_ERR "RBF: error reading bitmap block %d\n", mapblk);
			return;
		}
		lock_super(s);
		if (s_su_paranoid && rbf_find_next_zero_bit(bh->b_data, length, offset) < offset + nclus)
			printk(KERN_ERR "RBF: sector(s) not allocated in %ld-%ld\n", (long) sector, (long) sector + nsect);
		
		rbf_clear_bits(bh->b_data, offset, nclus);
		s_su_freesect += nsect;
		rbf_update_bitmap(s, bh);
		unlock_super(s);
		brelse(bh);
		
		count -= nclus;
		sector += nsect;
	}
	DPRINTK("out\n");
}


/*
 *	called by read_super
 *	superblock locked
 */
long rbf_free_blocks(struct super_block *s)
{
	int length, offset, start, end;
	long mapblk, sector = 0;
	int size, frag, ngap = 0, nfree = 0, largest_gap = 0;
	struct buffer_head *bh;
	
	while ((mapblk = rbf_bitmap_block(s, sector, &length, &offset)) != -1) {
		bh = sb_bread(s, mapblk);
		if (bh == 0) {
			printk(KERN_ERR "RBF: error reading bitmap block %ld\n", mapblk);
			goto next;
		}
		start = rbf_find_next_zero_bit(bh->b_data, length, offset);
		while (start < length) {
			end = rbf_find_next_one_bit(bh->b_data, length, start);
			ngap++;
			size = end - start;
			nfree += size;
			if (size > largest_gap)
				largest_gap = size;
			start = rbf_find_next_zero_bit(bh->b_data, length, end);
		}
		brelse(bh);
	next:
		sector += (length - offset) << s_su_clus_bits;
	}
	
	frag = (ngap * 200) / s_su_nclus;	/* a rough'n'ready estimate */
	
	if (s_su_verbose) {
		printk("RBF device %s free clusters %d/%d largest %d fragmentation %d%%\n",
				s->s_id, nfree, s_su_nclus, largest_gap, frag);
	}
	return nfree << s_su_clus_bits;	/* convert to sectors */
}

#ifdef RBF_HIGHZ_TABLE
char rbf_highz[256] = {
	8, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4,
	3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};
#endif

