/*
 *	linux/fs/rbf/super.c
 *	RBF File System - superblock operations.
 *	
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 *	
 * from
 * 
 *  fs/minix, Copyright (C) 1991, 1992 Linus Torvalds.
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"
#include <linux/blkdev.h>

/* #undef DEBUG */
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

#ifdef NODEBUG
#define RBF_STATUS	RBF_LEVEL
#else
#define RBF_STATUS	RBF_LEVEL " [DEBUG]"
#endif

/* Helper functions */
char *index(const char *s, int c)
{
  do {
    if (*s == c)
      return((char*) s);
  } while (*s++);
  return(0);
}

char *strtok(char *s1, const char *s2)
{
  static char *last_s1;
  register char *p, *q;
                                                                                
  if ((s1 == NULL && last_s1 == NULL) || s2 == NULL)
    return(NULL);
                                                                                
  p = s1 ? s1 : last_s1;
                                                                                
  while ( *p && index(s2, *p))
    p++;
                                                                                
  q = *p ? p : NULL;
                                                                                
  if ( q ) {
    while( *p && !index(s2, *p))
      p++;
                                                                                
    last_s1 = *p ? p + 1 : p ;
    *p = '\0';
  }
                                                                                
  return(q);
}

static struct kmem_cache *rbf_inode_cachep;

static struct inode *rbf_alloc_inode(struct super_block *sb)
{
	struct rbf_inode_info *ei;
	ei = kmem_cache_alloc(rbf_inode_cachep, GFP_KERNEL);
	if (!ei)
		return NULL;
	return &ei->vfs_inode;
}

static void rbf_destroy_inode(struct inode *inode)
{
	kmem_cache_free(rbf_inode_cachep, rbf_i(inode));
}

static void init_once(void *foo, struct kmem_cache * cachep,
		      unsigned long flags)
{
	struct rbf_inode_info *ei = (struct rbf_inode_info *) foo;

	if ((flags & (SLAB_CTOR_VERIFY | SLAB_CTOR_CONSTRUCTOR)) ==
	    SLAB_CTOR_CONSTRUCTOR)
		inode_init_once(&ei->vfs_inode);
}

static int init_inodecache(void)
{
	rbf_inode_cachep = kmem_cache_create("rbf_inode_cache",
					     sizeof(struct rbf_inode_info),
					     0, SLAB_HWCACHE_ALIGN|SLAB_RECLAIM_ACCOUNT,
					     init_once, NULL);
	if (rbf_inode_cachep == NULL)
		return -ENOMEM;
	return 0;
}

static void destroy_inodecache(void)
{
	kmem_cache_destroy(rbf_inode_cachep);
}

static void rbf_read_inode(struct inode *inode)
{
	unsigned long ino = inode->i_ino;
	struct rbf_inode *di;
	struct buffer_head *bh;
	struct super_block *s = inode->i_sb;
	int offset;
	
	DPRINTK("ino=%08lx\n", ino);

	if (ino != s_su_root && (ino < s_su_root || ino >= s_su_nsect || ino & (s_su_clus - 1))) {
		printk(KERN_ERR "Bad inode number %s:%08lx\n", s->s_id, ino);
		make_bad_inode(inode);
		return;
	}

	bh = rbf_bread(s, inode_adrs(s, ino), &offset);
	if (!bh) {
		printk(KERN_ERR "Unable to read inode %s:%08lx\n", s->s_id, ino);
		make_bad_inode(inode);
		return;
	}
	di = (struct rbf_inode *)(bh->b_data + offset);
	
/*	DPRINTK("%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
		((u8*)di)[0],
		((u8*)di)[1],
		((u8*)di)[2],
		((u8*)di)[3],
		((u8*)di)[4],
		((u8*)di)[5],
		((u8*)di)[6],
		((u8*)di)[7],
		((u8*)di)[8],
		((u8*)di)[9],
		((u8*)di)[10],
		((u8*)di)[11],
		((u8*)di)[12],
		((u8*)di)[13],
		((u8*)di)[14],
		((u8*)di)[15]);
*/		
	
	inode->i_blocks = rbf_scan_fd(inode, di);
	if (inode->i_blocks == 0) {
		make_bad_inode(inode);
		goto out;
	}

	inode->i_mode = (di->fd_att << 8 & 0400)
		|	(di->fd_att << 6 & 0200)
		|	(di->fd_att << 4 & 0100)
		|	(di->fd_att << 2 & 0040)			/* copy public attr to group */
		|	(di->fd_att      & 0020)
		|	(di->fd_att >> 2 & 0010)
		|	(di->fd_att >> 1 & 0004)
		|	(di->fd_att >> 3 & 0002)
		|	(di->fd_att >> 5 & 0001);
	
	if (di->fd_att & RBF_S_IFDIR) {
		inode->i_mode |= S_IFDIR;
		inode->i_mode |= (inode->i_mode >> 2) & S_IXUGO;	/* copy r to x for dirs */
		inode->i_op = s_su_operations->dir_inode_ops;
		inode->i_fop = s_su_operations->dir_operations;
	}
	else {
		inode->i_mode |= S_IFREG;
		inode->i_op = s_su_operations->file_inode_ops;
		inode->i_fop = s_su_operations->file_operations;
		i_mmu_private = inode->i_size;
	}
	inode->i_mapping->a_ops = s_su_operations->address_space_ops;
	
	inode->i_uid = s_su_uid ?: s_su_owner ? (di->fd_own[0] << 8) + di->fd_own[1] : di->fd_own[1];
	inode->i_gid = s_su_gid ?: s_su_owner ? current->fsgid : di->fd_own[0];
	
	inode->i_size = rbf_read(di->fd_fsize, 4);
	inode->i_atime.tv_sec = inode->i_mtime.tv_sec = inode->i_ctime.tv_sec = rbf_rdtime(di->fd_date, 5);
        inode->i_atime.tv_nsec = inode->i_mtime.tv_nsec = inode->i_ctime.tv_nsec = 0;
	i_iu_create = rbf_rdtime(di->fd_dcr, 3);
	
	// problem: real RBF doesn't maintain "correct" link count for dirs - what to do?
	//
	inode->i_nlink = di->fd_link;
	if (inode->i_nlink == 0)
		inode->i_nlink = (di->fd_att & RBF_S_IFDIR) ? 2 : 1;	// default
	
//	DPRINTK("mode=%04x size=%08lx\n", (int)inode->i_mode, (long)inode->i_size);
	
	i_iu_segnum = -1;
	i_iu_segend = 0;
	
out:	
	brelse(bh);
}

static int rbf_write_inode(struct inode *inode, int unused)
{
	unsigned long ino = inode->i_ino;
	struct rbf_inode *di;
	struct buffer_head *bh;
	struct super_block *s = inode->i_sb;
	int offset;
	int mode;

	DPRINTK("ino=%08lx\n", ino);

	if (ino != s_su_root && (ino < s_su_root || ino >= s_su_nsect || ino & (s_su_clus - 1))) {
		printk(KERN_ERR "Bad inode number %s:%08lx\n", s->s_id, ino);
		return -EIO;
	}

	lock_kernel();
	bh = rbf_bread(s, inode_adrs(s, ino), &offset);
	if (!bh) {
		printk(KERN_ERR "Unable to read inode %s:%08lx\n", s->s_id, ino);
		unlock_kernel();
		return -EIO;
	}
	di = (struct rbf_inode *)(bh->b_data + offset);
	
	mode = inode->i_mode | (inode->i_mode >> 3 & 7);	/* public = group | others */
	di->fd_att =	  (mode >> 8 & 1)
			| (mode >> 6 & 2)
			| (mode >> 4 & 4)
			| (mode << 1 & 0x08)
			| (mode << 3 & 0x10)
			| (mode << 5 & 0x20);
	if (inode->i_mode & S_IFDIR)
		di->fd_att |= RBF_S_IFDIR;
		
	/*
	 *	Hmmm, this doesn't really do what we want...
	 *	inodes will be updated even if owner hasn't been changed
	 *	...but only those that have been marked dirty!!!
	 */
	if (s_su_owner) {
		uid_t owner = s_su_setuid != -1 ? s_su_setuid : inode->i_uid;
		di->fd_own[0] = owner >> 8;
		di->fd_own[1] = owner;
	}
	else {
		di->fd_own[1] = s_su_setuid != -1 ? s_su_setuid : inode->i_uid;
		di->fd_own[0] = s_su_setgid != -1 ? s_su_setgid : inode->i_gid;
	}
	di->fd_link = inode->i_nlink;
	rbf_write(di->fd_fsize, 4, inode->i_size);
	rbf_wrtime(di->fd_date, 5, inode->i_mtime.tv_sec);
	rbf_wrtime(di->fd_dcr, 3, i_iu_create);
	
	rbf_update_buffer(IS_SYNC(inode), bh);
	brelse(bh);
	unlock_kernel();
	return 0;
}

static void rbf_delete_inode(struct inode *inode)
{
	unsigned long ino = inode->i_ino;
	struct super_block *s = inode->i_sb;

	DPRINTK("ino=%08lx\n", ino);

	if (ino != s_su_root && (ino < s_su_root || ino >= s_su_nsect || ino & (s_su_clus - 1))) {
		printk(KERN_ERR "Bad inode number %s:%08lx\n", s->s_id, ino);
		return;
	}
	
	truncate_inode_pages(&inode->i_data, 0);
	inode->i_size = 0;
	inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;

	rbf_contract(inode, 0);				/* release all segments */
	lock_kernel();
	rbf_deallocate(s, inode_adrs(s, ino), 1);	/* release FD sector */
	unlock_kernel();
	clear_inode(inode);
}

static void rbf_put_super(struct super_block *s)
{
	brelse(s_su_sbh);
}

static void rbf_write_super(struct super_block *s)
{
	lock_kernel();
	if (!(s->s_flags & MS_RDONLY))
		mark_buffer_dirty(s_su_sbh);
	s->s_dirt = 0;
	unlock_kernel();
}

static int rbf_statfs(struct dentry *dentry, struct kstatfs *buf)
{
        struct super_block *s = dentry->d_sb;

	lock_kernel();

	buf->f_type = s->s_magic;
	buf->f_bsize = s_su_sectsize;
	buf->f_blocks = s_su_nsect;
	buf->f_bfree = buf->f_bavail = s_su_freesect;
	buf->f_files = 0;
	buf->f_ffree = 0;
	/* buf->f_fsid.val[0] = kdev_t_to_nr(s->s_dev); */
	buf->f_namelen = RBF_NAMELEN;

	unlock_kernel();

	return 0;
}

static int rbf_parse_options(char *options, struct super_block *s)
{
	char *this_opt, *value;
	
	if (!options)
		goto out;

	for (this_opt = strtok(options, ","); this_opt; this_opt = strtok(0, ",")) {
		if ((value = strchr(this_opt, '=')) != 0)
			*value++ = 0;
		
		if (strcmp(this_opt, "debug") == 0 && !value) {
			s_su_debug = 1;
		}
		else if (strcmp(this_opt, "verbose") == 0 && !value) {
			s_su_verbose = 1;
		}
		else if (strcmp(this_opt, "quiet") == 0 && !value) {
			s_su_quiet = 1;
		}
		else if (strcmp(this_opt, "dealloc") == 0 && !value) {
			s_su_dealloc = 1;
		}
		else if (strcmp(this_opt, "keepcase") == 0 && !value) {
			s_su_keepcase = 1;
		}
		else if (strcmp(this_opt, "hardlinks") == 0 && !value) {
			s_su_hardlinks = 1;
		}
		else if (strcmp(this_opt, "owner") == 0 && !value) {
			s_su_owner = 1;
		}
		else if (strcmp(this_opt, "paranoid") == 0 && !value) {
			s_su_paranoid = 1;
		}
		else if (strcmp(this_opt, "nomagic") == 0 && !value) {
			s_su_nomagic = 1;
		}
		else if (strcmp(this_opt, "nonamecheck") == 0 && !value) {
			s_su_nonamecheck = 1;
		}
		else if (strcmp(this_opt, "sector") == 0 && value && *value) {
			s_su_sectsize = simple_strtoul(value, &value, 0);
			if (*value) return 0;
			if (s_su_sectsize < 256
			 || s_su_sectsize > 32768
			 || (s_su_sectsize & -s_su_sectsize) != s_su_sectsize)
				return 0;
		}
		else if (strcmp(this_opt, "sas") == 0 && value && *value) {
			s_su_minalloc = simple_strtoul(value, &value, 0);
			if (*value == 'k') {
				value++;
				s_su_minalloc_k = 1;
			}
			if (*value) return 0;
		}
		else if (strcmp(this_opt, "umask") == 0 && value && *value) {
			s_su_umask = simple_strtoul(value, &value, 8);
			if (*value) return 0;
		}
		else if (strcmp(this_opt, "uid") == 0 && value && *value) {
			s_su_uid = simple_strtoul(value, &value, 0);
			if (*value) return 0;
		}
		else if (strcmp(this_opt, "gid") == 0 && value && *value) {
			s_su_uid = simple_strtoul(value, &value, 0);
			if (*value) return 0;
		}
		else if (strcmp(this_opt, "setuid") == 0 && value && *value) {
			s_su_setuid = simple_strtoul(value, &value, 0);
			if (*value) return 0;
		}
		else if (strcmp(this_opt, "setgid") == 0 && value && *value) {
			s_su_setuid = simple_strtoul(value, &value, 0);
			if (*value) return 0;
		}
		else return 0;
	}
out:
	return 1;
}

static struct super_operations rbf_super_ops = {
        .alloc_inode	= rbf_alloc_inode,
        .destroy_inode	= rbf_destroy_inode,
	.read_inode	= rbf_read_inode,
	.write_inode	= rbf_write_inode,
	.delete_inode	= rbf_delete_inode,
	.put_super	= rbf_put_super,
	.write_super	= rbf_write_super,
	.statfs		= rbf_statfs
};

struct rbf_operations rbf_block_operations = {
	.dir_inode_ops =	&rbf_block_dir_inode_ops,
	.dir_operations =	&rbf_block_dir_operations,
	.file_inode_ops =	&rbf_file_inode_ops,
	.file_operations =	&rbf_block_file_operations,
	.address_space_ops =	&rbf_block_address_space_ops
};

struct rbf_operations rbf_page_operations = {
	.dir_inode_ops =	&rbf_page_dir_inode_ops,
	.dir_operations =	&rbf_page_dir_operations,
	.file_inode_ops =	&rbf_file_inode_ops,
	.file_operations =	&rbf_page_file_operations,
	.address_space_ops =	&rbf_page_address_space_ops
};

/*
 * RBF file system blocks are referred to as "sectors" to avoid confusion
 * and may be from 256 bytes to 32k in size.
 * 
 * In the normal case the buffer block size is equal to the sector size.
 * 
 * For sector size < device hard-blocksize we must use partial block I/O.
 * In this mode page cache I/O operations cannot be used so we switch to
 * buffer-based I/O operations; mmap is not available.
 * This will always be the case for RBF filesystems with 256 byte sectors.
 * 
 * For sector size > PAGE_CACHE_SIZE (normally 4096 bytes) the buffer block
 * size is set to the maximum of PAGE_CACHE_SIZE and a buffer will hold a
 * partial sector.
 * The only caveat in this mode is that the current implementation only
 * reads and writes FD segments in the first block. An error will be
 * returned on encountering a file with segments beyond the first block;
 * this should not normally be a problem since a block size of 4096 allows
 * up to 816 segments.
 */
static int rbf_read_super(struct super_block *s, void *data, int silent)
{
	struct buffer_head *bh;
	struct inode *inode;
	struct rbf_super_block *rsb;
	struct rbf_sb_info *rsi;
	int blocksize;
	int sectsize;
	int mapsize;

	rsi = kzalloc(sizeof(struct rbf_sb_info), GFP_KERNEL);
	if (!rsi)
		return -ENOMEM;
	s->s_fs_info = rsi;

	s_su_setuid = -1;
	s_su_setgid = -1;
	if (!rbf_parse_options((char *)data, s))
		goto error;
	s_su_umask |= 07000;				/* always disable special file attrs */

	blocksize = max(bdev_hardsect_size(s->s_bdev), RBF_BSIZE);
	DPRINTK("blocksize=%d, max(%d,%d)\n",
		blocksize, bdev_hardsect_size(s->s_bdev), RBF_BSIZE);
	sb_set_blocksize(s, blocksize);
	s->s_blocksize = blocksize;
	s->s_blocksize_bits = ffs(blocksize) - 1;

	bh = sb_bread(s, 0);
	if (bh == 0) {
                printk("rbf: unable to read the superblock\n");
		goto out;
	}
	rsb = (struct rbf_super_block *)bh->b_data;

	s->s_blocksize = blocksize;
	s->s_blocksize_bits = ffs(blocksize) - 1;
	
	s_su_nsect = rbf_read(rsb->dd_tot, 3);
	s_su_clus = be16_to_cpu(rsb->dd_bit);
	s_su_clus_bits = ffs(s_su_clus) - 1;
	s_su_root = rbf_read(rsb->dd_dir, 3);
	s_su_mapblk = be32_to_cpu(rsb->dd_maplsn);
	if (s_su_mapblk == 0)
		s_su_mapblk = 1;			/* default for old media */

	mapsize = be16_to_cpu(rsb->dd_map);
	sectsize = be16_to_cpu(rsb->dd_lsnsize);
	
	/*
	 *	Perform a few sanity checks on DD sector
	 */
	if ((sectsize & -sectsize) != sectsize || (sectsize & 0xff))
		goto notrbf;
	if (s_su_clus == 0 || (s_su_clus & -s_su_clus) != s_su_clus)
		goto notrbf;
	if ((mapsize - 1) * 8 * s_su_clus > s_su_nsect)
		goto notrbf;
	
	s_su_nclus = min((long)mapsize * 8, s_su_nsect >> s_su_clus_bits);
	s_su_nsect = s_su_nclus << s_su_clus_bits;
	
	if (s_su_root >= s_su_nsect)
		goto notrbf;
	if (s_su_mapblk < 0 || s_su_mapblk >= s_su_nsect)
		goto notrbf;
	
	s->s_magic = be32_to_cpu(rsb->dd_magic);
	if (s->s_magic == 0 && s_su_nomagic) {		/* for old media w/o sync word */
		s->s_magic = RBF_MAGIC;
		if (s_su_sectsize == 0)
			sectsize = RBF_BSIZE;		/* default to 256 unless option overrides */
	}
	
	if (s->s_magic != RBF_MAGIC) {
	notrbf:
		if (!silent)
			printk(KERN_ERR "No RBF filesystem on %s (magic=%08lx)\n", 
				s->s_id, s->s_magic);
		goto out;
	}
	
	if (sectsize == 0)
		sectsize = s_su_sectsize;		/* use specified size if not explicitly set */
	if (sectsize == 0)
		sectsize = blocksize;			/* default if not specified at all */
	s_su_sectsize = sectsize;
	
	DPRINTK("nsect=%08lx mapsize=%04x clus=%04x root=%ld maplsn=%ld\n",
			s_su_nsect, mapsize, s_su_clus, s_su_root, s_su_mapblk);
	
	s_su_sectsize = sectsize;
	s_su_sectsize_bits = ffs(sectsize) - 1;
	s_su_operations = &rbf_page_operations;
	
	if (s_su_minalloc_k)
		s_su_minalloc = (((s_su_minalloc << 2) - 1) >> (s_su_sectsize_bits - 8)) + 1;
	
	if (sectsize < blocksize) {
		/*
		 *	Sector is smaller than disk block...
		 *	Set up for partial block I/O.
		 *	Page cache cannot be used, so switch to block-based I/O operations
		 */
		if (!s_su_quiet)
			printk("RBF device %s *warning* sector size %d mmap unavailable\n",
					s->s_id, sectsize);
		s_su_part_shift = s->s_blocksize_bits - s_su_sectsize_bits;
		s_su_part_mask = (1 << s_su_part_shift) - 1;
		s_su_mapoffs = (s_su_mapblk & s_su_part_mask) << (s_su_sectsize_bits + 3);
		s_su_mapblk >>= s_su_part_shift;
		s_su_operations = &rbf_block_operations;
	}
	else if (sectsize <= PAGE_CACHE_SIZE) {
		/*
		 *	The usual case (hopefully)
		 */
		blocksize = sectsize;
	}
	else {
		/*
		 *	Sector is larger than page size.
		 *	This is not fully supported yet (in inode.c)
		 *	Only the first 4096 bytes of the FD may contain segments
		 */
		if (!s_su_quiet)
			printk("RBF device %s *warning* sector size %d not fully compliant\n",
					s->s_id, sectsize);
		blocksize = PAGE_CACHE_SIZE;
		s_su_sect_shift = s_su_sectsize_bits - PAGE_CACHE_SHIFT;
		s_su_sect_mask = (1 << s_su_sect_shift) - 1;
		s_su_mapblk <<= s_su_sect_shift;
	}

	if (blocksize != s->s_blocksize) {
		bforget(bh);
		DPRINTK("setting block size to %d\n", blocksize);
		sb_set_blocksize(s, blocksize);
		s->s_blocksize = blocksize;
		s->s_blocksize_bits = ffs(blocksize) - 1;
		bh = sb_bread(s, 0);
		if (bh == 0)
			goto out;
		rsb = (struct rbf_super_block *)bh->b_data;
	}
	s_su_sbh = bh;
	s_su_rbf_sb = rsb;
	s_su_maxseg = ((s_su_sectsize >> s_su_sect_shift) - 16) / 5;

	DPRINTK("nsect=%ld (%08lx) nclus=%d maxseg=%d mapoffs=%d\n",
			s_su_nsect, s_su_nsect, s_su_nclus, s_su_maxseg, s_su_mapoffs);
	
	DPRINTK("blocksize=%d sectsize=%d sectshift=%d sectmask=%d partshift=%d partmask=%d\n",
			(int) s->s_blocksize, (int) s_su_sectsize, s_su_sect_shift, s_su_sect_mask,
			s_su_part_shift, s_su_part_mask);
	
	if (s_su_verbose) {
		printk("RBF device %s sector size %ld cluster size %d sas %d\n",
				s->s_id, s_su_sectsize, s_su_clus, s_su_minalloc);
	}
	s_su_freesect = rbf_free_blocks(s);
	s->s_op = &rbf_super_ops;
	inode = iget(s, s_su_root);
	if (inode == 0)
		goto out;
	
	s->s_root = d_alloc_root(inode);
	if (s->s_root == 0) {
		iput(inode);
		goto out;
	}
	
#ifndef CONFIG_RBF_FS_WRITE
	if (!(s->s_flags & MS_RDONLY)) {
		if (!s_su_quiet)
			printk("RBF forcing read-only mode on %s\n", s->s_id);
		s->s_flags |= MS_RDONLY;
	}
#endif	

	return 0;

out:
	brelse(bh);
error:
	kfree(rsi);
	s->s_fs_info = NULL;
	return -EINVAL;
}

static int rbf_get_sb(struct file_system_type *fs_type, int flags,
	const char *dev_name, void *data, struct vfsmount *mnt)
{
	return get_sb_bdev(fs_type, flags, dev_name, data, rbf_read_super, mnt);
}

static struct file_system_type rbf_fs_type = {
	.owner		= THIS_MODULE,
	.name		= "rbf",
	.get_sb		= rbf_get_sb,
	.kill_sb	= kill_block_super,
	.fs_flags	= FS_REQUIRES_DEV,
};

static int __init init_rbf_fs(void)
{
	int err;

	err = init_inodecache();
	if (err)
		return err;

	err = register_filesystem(&rbf_fs_type);
	if (err) {
		destroy_inodecache();
		return err;
	}

	printk("RBF Filesystem Version %d.%02d "RBF_STATUS"\n", RBF_VERSION, RBF_REVISION);
	return 0;
}

static void __exit exit_rbf_fs(void)
{
	unregister_filesystem(&rbf_fs_type);
	destroy_inodecache();
}

module_init(init_rbf_fs)
module_exit(exit_rbf_fs)

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andrew Cannon <ajc@gmx.net>");
MODULE_DESCRIPTION("Microware/OS-9 RBF filesystem for Linux");
